# Experimental class for series, floats with lower and upper tolerance bounds

import numpy as np
import pandas as pd

default_tolerances = {
    "default": {(0, 1e3): 1, (1e3, 1e6): 2, (1e6, 1e8): 3, (1e8, np.inf): 4}
}


def _tol(value, tolerance_def):
    if isinstance(value, Interval):
        return value
    else:
        if tolerance_def is None:
            return value
        else:
            for ((start, end)), decimals in tolerance_def.items():
                if abs(value) >= start and abs(value) < end:
                    return 0.5 * 10 ** (decimals)
            raise Exception(
                "tolerance definition not found for value "
                + str(value)
                + " in "
                + tolerance_def
            )


class Interval(tuple):
    def __init__(self, x):
        """ """
        return tuple.__init__(x)

    def __add__(self, other):
        """ """
        if isinstance(other, Interval):
            self = Interval(
                [
                    self[0] + other[0],
                    self[1] + other[1],
                ],
            )
            return self
        else:
            self = Interval(
                [
                    self[0] + other,
                    self[1] + other,
                ]
            )
            return self

    def __radd__(self, other):
        """ """
        return self.__add__(other)

    def __sub__(self, other):
        """ """
        if isinstance(other, Interval):
            return Interval(
                [
                    self[0] - other[0],
                    self[1] - other[1],
                ]
            )
        else:
            return Interval(
                [
                    self[0] - other,
                    self[1] - other,
                ]
            )

    def __mul__(self, other):
        if isinstance(other, Interval):
            return Interval(
                [
                    min(
                        [
                            self[0] * other[0],
                            self[1] * other[0],
                            self[0] * other[1],
                            self[1] * other[1],
                        ]
                    ),
                    max(
                        [
                            self[0] * other[0],
                            self[1] * other[0],
                            self[0] * other[1],
                            self[1] * other[1],
                        ]
                    ),
                ]
            )
        else:
            return Interval(
                [
                    min(
                        [
                            self[0] * other,
                            self[1] * other,
                        ]
                    ),
                    max(
                        [
                            self[0] * other,
                            self[1] * other,
                        ]
                    ),
                ]
            )

    def __rmul__(self, other):
        """ """
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Interval):
            return Interval(
                [
                    min(
                        [
                            self[0] / other[0],
                            self[1] / other[0],
                            self[0] / other[1],
                            self[1] / other[1],
                        ]
                    ),
                    max(
                        [
                            self[0] / other[0],
                            self[1] / other[0],
                            self[0] / other[1],
                            self[1] / other[1],
                        ]
                    ),
                ]
            )
        else:
            return Interval(
                [
                    min(
                        [
                            self[0] / other,
                            self[1] / other,
                        ]
                    ),
                    max(
                        [
                            self[0] / other,
                            self[1] / other,
                        ]
                    ),
                ]
            )

    def __floordiv__(self, other):
        if isinstance(other, Interval):
            return Interval(
                [
                    min(
                        [
                            self[0] // other[0],
                            self[1] // other[0],
                            self[0] // other[1],
                            self[1] // other[1],
                        ]
                    ),
                    max(
                        [
                            self[0] // other[0],
                            self[1] // other[0],
                            self[0] // other[1],
                            self[1] // other[1],
                        ]
                    ),
                ]
            )
        else:
            return Interval(
                [
                    min(
                        [
                            self[0] // other,
                            self[1] // other,
                        ]
                    ),
                    max(
                        [
                            self[0] // other,
                            self[1] // other,
                        ]
                    ),
                ]
            )

    def __pow__(self, other):
        if isinstance(other, Interval):
            return Interval(
                [
                    min(
                        [
                            self[0] ** other[0],
                            self[1] ** other[0],
                            self[0] ** other[1],
                            self[1] ** other[1],
                        ]
                    ),
                    max(
                        [
                            self[0] ** other[0],
                            self[1] ** other[0],
                            self[0] ** other[1],
                            self[1] ** other[1],
                        ]
                    ),
                ]
            )
        else:
            return Interval(
                [
                    min(
                        [
                            self[0] ** other,
                            self[1] ** other,
                        ]
                    ),
                    max(
                        [
                            self[0] ** other,
                            self[1] ** other,
                        ]
                    ),
                ]
            )

    def __eq__(self, other):
        return (self[1] >= other[0]) & (self[0] <= other[1])

    def __ne__(self, other):
        return ~((self[1] >= other[0]) & (self[0] <= other[1]))

    def __ge__(self, other):
        return self[1] >= other[0]

    def __gt__(self, other):
        return self[0] > other[1]

    def __le__(self, other):
        return self[0] <= other[1]

    def __lt__(self, other):
        return self[1] < other[0]

    def __abs__(self):
        return Interval(
            [
                # min of abs(upper) and abs(lower), except if lower < 0 and upper > 0
                # (then the result should be 0)
                min(
                    [
                        abs(self[0]),
                        abs(self[1]),
                    ]
                )
                if not (self[0] < 0 and self[1] > 0)
                else 0,
                # max of abs(upper) and abs(lower)
                max(
                    [
                        abs(self[0]),
                        abs(self[1]),
                    ]
                ),
            ]
        )

    def __pos__(self):
        return Interval([self[0], self[1]])

    def __neg__(self):
        return Interval([-self[1], -self[0]])

    # def __mod__(self, other):
    #     pass

    # def __divmod__(self, other):
    #     pass

    # def __ceil__(self, other):
    #     pass

    # def __floor__(self, other):
    #     pass

    # def __round__(self, other):
    #     pass


class SeriesWithTolerance(pd.Series):
    # add property to retain original data
    _metadata = ["original_data"]

    def __init__(self, *args, **kwargs):
        """ """
        # retain the original data within the object
        self.original_data = kwargs.get("data", *args)
        # change values to intervals before creating the series
        kwargs["data"] = [
            Interval(
                [
                    value - _tol(value, default_tolerances["default"]),
                    value + _tol(value, default_tolerances["default"]),
                ]
            )
            if not isinstance(value, Interval)
            else value
            for value in kwargs.get("data", *args)
        ]
        super().__init__(**kwargs)

    def __eq__(self, other):
        """ """
        res = self.values == other.values
        return res

    def __ne__(self, other):
        """ """
        return self.values != other.values

    def __ge__(self, other):
        """ """
        return self.values >= other.values

    def __gt__(self, other):
        """ """
        return self.values > other.values

    def __le__(self, other):
        """ """
        return self.values <= other.values

    def __lt__(self, other):
        """ """
        return self.values < other.values

    @property
    def _constructor(self):
        return SeriesWithTolerance

    @property
    def _constructor_expanddim(self):
        return DataFrameWithTolerance


class DataFrameWithTolerance(pd.DataFrame):
    # add property to retain original data
    _metadata = ["original_data"]

    def __init__(self, *args, **kwargs):
        """ """
        # retain the original data within the object
        self.original_data = kwargs.get("data", *args)
        # change values to intervals before creating the series
        kwargs["data"] = [
            [
                Interval(
                    [
                        value - _tol(value, default_tolerances["default"]),
                        value + _tol(value, default_tolerances["default"]),
                    ]
                )
                if not isinstance(value, Interval)
                else value
                for value in column
            ]
            for column in kwargs.get("data", *args)
        ]
        super().__init__(**kwargs)

    @property
    def _constructor(self):
        return DataFrameWithTolerance

    @property
    def _constructor_sliced(self):
        return SeriesWithTolerance



"""Main module."""

import pandas as pd
import logging
import re
import abc
import numpy as np
from sklearn.tree import _tree, DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.base import is_classifier, is_regressor


def generate_substitutions(
    df: pd.DataFrame = None,
    column_value: tuple = None,
    value_regex: str = None,
):
    """
    Util function to retrieve values of dataframe satisfying a regex
    """
    if df is not None:
        column_regex, value_regex = column_value

        compiled_column_regex = re.compile(column_regex[2:-2])
        columns_found = [
            col
            for col in df.columns
            if compiled_column_regex.fullmatch(col) is not None
        ]
        # if there is a value_regex then we compile it in advance
        if value_regex is not None and value_regex[0] != "{":
            compiled_value_regex = re.compile(value_regex[1:-1])
        else:
            compiled_value_regex = None
        for column in columns_found:
            if compiled_value_regex is not None:
                for value in df[column].unique():
                    if isinstance(value, str):
                        try:
                            r = compiled_value_regex.fullmatch(value)
                        except Exception as e:
                            logging.error(
                                "Error evaluating regex: "
                                + value_regex[1:-1]
                                + ", "
                                + str(e)
                            )
                        if r is not None:
                            yield (column, value, r.groups())
            else:
                yield (column, None, None)


def tree_to_expressions(tree, features, target):
    """
    Util function to derive rules from a decision tree (classifier or regressor)
    """
    tree_ = tree.tree_
    feature_name = [
        features[i] if i != _tree.TREE_UNDEFINED else "undefined!"
        for i in tree_.feature
    ]

    def evaluate_node(node, expr, solutions):
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            # not a leaf, evaluate left and right path
            name = feature_name[node]
            threshold = tree_.threshold[node]
            evaluate_node(
                tree_.children_left[node],
                expr
                + (" & " if expr != "" else "")
                + '({"'
                + name
                + '"} <= '
                + str(threshold)
                + ")",
                solutions,
            )
            evaluate_node(
                tree_.children_right[node],
                expr
                + (" & " if expr != "" else "")
                + '({"'
                + name
                + '"} > '
                + str(threshold)
                + ")",
                solutions,
            )
        else:
            # leaf node so add complete path to solutions
            value = tree_.value[node][0]
            if isinstance(tree, DecisionTreeClassifier):
                class_name = np.argmax(value)
                if isinstance(tree.classes_[class_name], str):
                    solutions.add(
                        "if ("
                        + expr
                        + ') then ({"'
                        + target
                        + '"} == "'
                        + tree.classes_[class_name]
                        + '")'
                    )
                else:
                    solutions.add(
                        "if ("
                        + expr
                        + ') then ({"'
                        + target
                        + '"} == '
                        + str(tree.classes_[class_name])
                        + ")"
                    )
            elif isinstance(tree, DecisionTreeRegressor):
                solutions.add(
                    "if ("
                    + expr
                    + ') then ({"'
                    + target
                    + '"} == '
                    + str(value[0])
                    + ")"
                )
            else:
                logging.error("Unknown classifier or regressor")

    solutions = set()
    evaluate_node(0, "", solutions)
    return solutions


def fit_ensemble_and_extract_expressions(
    df: pd.DataFrame = None,
    target: str = None,
    estimator: abc.ABCMeta = None,
    base: abc.ABCMeta = None,
    random_state: int = 0,
    max_depth: int = 2,
    n_estimators: int = 10,
    min_samples_split: int = 2,
    min_samples_leaf: int = 1,
    min_weight_fraction_leaf: float = 0.0,
    sample_weight: list = None,
):
    features = [col for col in df.columns if col != target]
    X = df[features]
    Y = df[[target]].values.ravel()
    target_dtype = df.dtypes[df.columns.get_loc(target)]

    if estimator is None:
        if pd.api.types.is_float_dtype(target_dtype):
            estimator = AdaBoostRegressor
        elif pd.api.types.is_integer_dtype(target_dtype):
            estimator = AdaBoostClassifier
    else:
        if pd.api.types.is_float_dtype(target_dtype) and not is_regressor(estimator):
            logging.error("target has float type data and estimator is not a regressor")
        elif pd.api.types.is_integer_dtype(target_dtype) and not is_classifier(
            estimator
        ):
            logging.error(
                "target has integer type data and estimator is not a classifier"
            )

    if base is None:
        if pd.api.types.is_float_dtype(target_dtype):
            base = DecisionTreeRegressor
        elif pd.api.types.is_integer_dtype(target_dtype):
            base = DecisionTreeClassifier
    else:
        if pd.api.types.is_float_dtype(target_dtype) and not is_regressor(base):
            logging.error("target has float type data and base is not a regressor")
        elif pd.api.types.is_integer_dtype(target_dtype) and not is_classifier(base):
            logging.error("target has integer type data and base is not a classifier")

    regressor = estimator(
        base_estimator=base(
            random_state=random_state,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
        ),
        n_estimators=n_estimators,
        random_state=random_state,
    )
    regressor = regressor.fit(
        X,
        Y,
        sample_weight=sample_weight,
    )
    ensemble_expressions = [
        tree_to_expressions(estimator, features, target)
        for estimator in regressor.estimators_
    ]
    return ensemble_expressions


def fit_dataframe_to_ensemble(
    df: pd.DataFrame = None,
    random_state: int = 0,
    max_depth: int = 1,
    n_estimators: int = 10,
    min_samples_split: int = 2,
    min_samples_leaf: int = 1,
    min_weight_fraction_leaf: float = 0.0,
):
    """
    fit and extract from an ensemble
    """
    solutions = set()
    for target in df.columns:
        ensemble_expressions = fit_ensemble_and_extract_expressions(
            df=df,
            target=target,
            random_state=random_state,
            max_depth=max_depth,
            n_estimators=n_estimators,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
        )

        for expressions in ensemble_expressions:
            for sol in expressions:
                solutions.add(sol)

    return solutions


def is_column(s):
    """
    Check if a given string is formatted as a column reference.

    This function checks if a string is formatted as a column reference,
    which typically consists of double curly braces {""} enclosing a
    column name.

    Args:
        s (str): The string to be checked.

    Returns:
        bool: True if the string is formatted as a column reference,
        False otherwise.

    Example:
        is_column('{"A"}')
            True

        is_column('{"B"}')
            True

        is_column("Not a column reference")
            False
    """
    return len(s) > 4 and (
        (s[:2] == '{"' and s[-1:] == "}") or (s[:2] == "{'" and s[-1:] == "}")
    )


def is_string(s):
    """
    Check if a given string is enclosed in single or double quotes.

    This function checks if a string is enclosed in single ('') or
    double ("") quotes, indicating that it is a string literal.

    Args:
        s (str): The string to be checked.

    Returns:
        bool: True if the string is enclosed in quotes, False otherwise.

    Example:
        is_string('"life"')
            True

        is_string('{"A"}')
            False

        is_string('""')
            True
    """
    return len(s) > 1 and (
        (s[:1] == '"' and s[-1:] == '"') or (s[:1] == "'" and s[-1:] == "'")
    )


def is_number(s):
    """
    Check if a given string is a number

    This function checks if a string is a number.

    Args:
        s (str): The string to be checked.

    Returns:
        bool: True if the string is a number

    Example:
        is_number('1.2')
            True

        is_number('{"A"}')
            False

    """
    if isinstance(s, str):
        pattern = r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?"
        match = re.match(pattern, s)
        return bool(match)
    else:
        return False


def flatten(expression):
    """
    Recursively flatten a nested expression and return it as a string.

    This function takes an expression, which can be a nested list of strings
    or a single string, and recursively flattens it into a single string
    enclosed in parentheses.

    Args:
        expression (str or list): The expression to be flattened.

    Returns:
        str: The flattened expression as a string enclosed in parentheses.

    Example:
        expression = ["A", ["B", ["C", "D"]]]

        result = ruleminer.flatten(expression)

        print(result)

            "(A(B(CD)))"
    """
    if isinstance(expression, str):
        return expression
    else:
        res = "".join([flatten(item) for item in expression])
        return res



"""Parser module."""

import pyparsing

# a class-level static method to enable a memorizing performance enhancement
pyparsing.ParserElement.enable_packrat()

_lpar = pyparsing.Literal("(")
_rpar = pyparsing.Literal(")")
_lbra = pyparsing.Literal("[")
_rbra = pyparsing.Literal("]")
_sep = pyparsing.Literal(",")
_quote = pyparsing.Literal('"')
_number = pyparsing.Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
_timedate_functions = pyparsing.one_of(
    "day_name \
    month_name \
    days_in_month \
    daysinmonth \
    is_leap_year \
    is_year_end \
    dayofweek \
    weekofyear \
    weekday \
    week \
    is_month_end \
    is_month_start \
    is_year_start \
    is_quarter_end \
    is_quarter_start \
    day \
    month \
    quarter \
    year",
    caseless=True,
)
_timedelta_functions = pyparsing.one_of("days months years", caseless=True)

_function = (
    pyparsing.one_of(
        "min \
    max \
    abs \
    quantile \
    sum \
    substr \
    split \
    count \
    sumif \
    countif \
    mean \
    std \
    exact \
    corr \
    round \
    floor \
    ceil \
    table",
        caseless=True,
    )
    | _timedelta_functions
    | _timedate_functions
)
_for = pyparsing.one_of("for", caseless=True)
_in = pyparsing.one_of("in", caseless=True)
_empty = pyparsing.one_of(["None", '""', "pd.NA", "np.nan"])
_list_comprehension_var = pyparsing.Word(pyparsing.alphas)
_string = (
    pyparsing.srange(r"[a-zA-Z0-9_.,:;<>*=+-/?|@#$%^&\[\]{}\(\)\\']")
    + " "
    + "\x01"
    + "\x02"
    + "\x03"
    # + pyparsing.pyparsing_unicode.Greek.alphas
    # + pyparsing.pyparsing_unicode.Greek.alphanums
)
_quoted_string = pyparsing.Combine(_quote + pyparsing.Word(_string) + _quote)
_column = pyparsing.Combine("{" + _quote + pyparsing.Word(_string) + _quote + "}")
_addop = pyparsing.one_of("+ -")
_multop = pyparsing.one_of("* /")
_expop = pyparsing.Literal("**")
_compa_op = (
    pyparsing.one_of(">= > <= < != == in between match contains", caseless=True)
    | pyparsing.CaselessLiteral("not in")
    | pyparsing.CaselessLiteral("not between")
    | pyparsing.CaselessLiteral("not match")
    | pyparsing.CaselessLiteral("not contains")
)

base_element = _quoted_string | _column | _number | _empty | _list_comprehension_var
math_expression = pyparsing.Forward()
condition_expression = pyparsing.infixNotation(
    (math_expression + _compa_op + math_expression),
    [
        (
            pyparsing.one_of(["not", "~"], caseless=True),
            1,
            pyparsing.opAssoc.RIGHT,
        ),
        (
            pyparsing.one_of(["and", "&"], caseless=True),
            2,
            pyparsing.opAssoc.LEFT,
        ),
        (
            pyparsing.one_of(["or", "|"], caseless=True),
            2,
            pyparsing.opAssoc.LEFT,
        ),
    ],
    lpar=pyparsing.Literal("("),
    rpar=pyparsing.Literal(")"),
)
param = pyparsing.Forward()
list_expression = pyparsing.Group(_lbra + param + (_sep + param)[...] + _rbra)
list_comprehension_expression = pyparsing.Group(
    _lbra
    + pyparsing.Group(param)
    + _for
    + _list_comprehension_var
    + _in
    + list_expression
    + _rbra
)
param <<= (
    list_comprehension_expression
    | list_expression
    | condition_expression
    | math_expression
)
params = param + (_sep + param)[...]
function = pyparsing.Group(_function + pyparsing.Group(_lpar + params + _rpar))
atom = (
    function
    | list_expression
    | base_element
    | pyparsing.Group(_lpar + math_expression + _rpar)
)
factor = pyparsing.Group(atom + pyparsing.OneOrMore(_expop + atom)) | atom
term = pyparsing.Group(factor + pyparsing.OneOrMore(_multop + factor)) | factor
math_expression <<= term + pyparsing.ZeroOrMore(_addop + term)

################################################################################
# definition of a rule expression
################################################################################
_if_then = (
    "if" + condition_expression + "then" + condition_expression
    | "IF" + condition_expression + "THEN" + condition_expression
)
rule_expression = (
    _if_then
    | "if () then " + condition_expression
    | "IF () THEN " + condition_expression
    | condition_expression
)
